<?php

namespace SuperRhino\Utils\MQClient;

use SuperRhino\Utils\MQClient\Exception\MQException;
use Exception;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\GuzzleException;
use GuzzleHttp\Psr7\Request;
use SuperRhino\Utils\GuzzleClient\GuzzleClient;
use SuperRhino\Utils\Logger\Log;
use XMLReader;
use XMLWriter;

/**
 * Class MQClient
 * @author Vitech
 * @package App\Util\AliyunMQ
 */
class MQClient {

    protected $endPoint;
    protected $accessId;
    protected $accessKey;
    protected $instanceId;

    protected $client;
    protected $requestTimeout = 35;
    protected $connectTimeout = 3;

    /**
     * MQClient constructor.
     * @param string $endPoint Endpoint URL
     * @param string $accessId
     * @param string $accessKey
     * @param string $instanceId
     */
    public function __construct(
        string $endPoint,
        string $accessId,
        string $accessKey,
        string $instanceId
    ) {
        $this->endPoint = $endPoint;
        $this->accessId = $accessId;
        $this->accessKey = $accessKey;
        $this->instanceId = $instanceId;

        $this->client = GuzzleClient::create([
            'base_uri' => $endPoint,
            'defaults' => [
                'headers' => [
                    'Host' => $endPoint
                ],
            ]
        ]);
    }

    /**
     * 生产消息
     * @param string $topic
     * @param string $messageBody
     * @param array $config <p>
     * @see Constants 一些消息配置，
     * MESSAGE_PROPERTIES_MSG_KEY : 设置消息KEY，如果没有设置，则消息的KEY为RequestId
     * MESSAGE_PROPERTIES_TIMER_KEY : 定时消息，单位毫秒（ms），在指定时间戳（当前时间之后）进行投递。
     * MESSAGE_PROPERTIES_TRANS_CHECK_KEY : 在消息属性中添加第一次消息回查的最快时间，单位秒，并且表征这是一条事务消息
     * </p>
     * @return string
     */
    public function produce(string $topic, string $messageBody, array $config = []) {
        return $this->sendProduceRequest($topic, $messageBody, $config);
    }

    /**
     * 消费消息，注意如果没有消息可以消费，已经做过预处理会返回空数组而不会像官方文档所述报错
     * @oaram string $topic
     * @param string $groupId 消费者id
     * @param int $numOfMessages 获取消息数量
     * @param int $waitSeconds <p>
     *     队列无消息可消费时的等待时间。如果传-1时不会等待
     *     比如传了10秒，那么在队列无消息时，请求会等待最多10秒钟，直到队列中插入了消息就会直接返回
     *     取值范围为-1 ~ 30
     * </p>
     * @return Message[]
     * @throws MQException
     */
    public function consume(string $topic, string $groupId, int $numOfMessages, int $waitSeconds = -1) {
        if ($numOfMessages < 0 || $numOfMessages > 16) {
            throw new MQException(400, 'numOfMessages should be 1~16');
        }
        if ($waitSeconds > 30) {
            throw new MQException(400, 'waitSeconds should be less than 30');
        }

        return $this->sendConsumeRequest($topic, $numOfMessages, $groupId, $waitSeconds);
    }

    /**
     * 确认对应receiptHandle的消息已经被消费
     * @param string $groupId 消费者
     * @param string[] $receiptHandles receiptHandle数组，从Message::getMessage中取出
     * @return bool
     * @see Message::$receiptHandle
     * @oaram string $topic
     */
    public function ack(string $topic, string $groupId, array $receiptHandles) {
        return $this->sendAckRequest($topic, $receiptHandles, $groupId);
    }

    /**
     * 确认对应消息数组的消息已经被消费
     * @oaram string $topic
     * @param string $groupId 消费者
     * @param Message[] $messages 从consume中取出的消息数组
     * @return bool
     */
    public function ackMessages(string $topic, string $groupId, array $messages) {
        $receiptHandles = array_map(function ($message) {
            return $message->receiptHandle;
        }, $messages);

        return $this->ack($topic, $groupId, $receiptHandles);
    }

    //生产
    private function sendProduceRequest(string $topic, string $messageBody, array $config = []) {
        $body = $this->generateProduceBody($messageBody, $config);
        $method = 'POST';
        $queryString = $this->generateProduceQueryString();

        $resourcePath = 'topics/' . $topic . '/messages';
        $headers = $this->addRequiredHeaders($resourcePath, $body, $method, $queryString);
        $request = new Request($method, $resourcePath, $headers);
        try {
            $response = $this->client->send($request, [
                'query' => $queryString,
                'body' => $body,
                'timeout' => $this->requestTimeout,
                'connect_timeout' => $this->connectTimeout
            ]);
            $statusCode = $response->getStatusCode();
            $content = $response->getBody();
            
            if ($statusCode == 201) {
                return $this->parseProduceXml($content);
            } else {
                //Log::error('MQ Produce Error ' . $statusCode, ["error" => '']);
                throw new MQException($statusCode, 'MQ Consume Error');
            }
        } catch (GuzzleException $exception) {
            //Log::error('MQ Produce Network Error', ["error" => $exception->getMessage()]);
            throw new MQException(500, 'MQ Produce Network Error');
        }
    }

    private function generateProduceQueryString() {
        if ($this->instanceId != null && $this->instanceId != "") {
            return http_build_query(array("ns" => $this->instanceId));
        }
        return NULL;
    }

    private function generateProduceBody(string $messageBody, array $properties = []) {
        $xmlWriter = new XMLWriter;
        $xmlWriter->openMemory();
        $xmlWriter->startDocument("1.0", "UTF-8");
        $xmlWriter->startElementNS(NULL, "Message", Constants::XML_NAMESPACE);
        if ($messageBody != NULL) {
            $xmlWriter->writeElement(Constants::MESSAGE_BODY, $messageBody);
        }
//        if ($messageTag !== NULL)
//        {
//            $xmlWriter->writeElement(Constants::MESSAGE_TAG, $messageTag);
//        }
        if ($properties !== NULL && sizeof($properties) > 0) {
            $this->checkPropValid($properties);
            $xmlWriter->writeElement(Constants::MESSAGE_PROPERTIES,
                implode("|", array_map(function ($v, $k) {
                    return $k . ":" . $v;
                }, $properties, array_keys($properties))));
        }
        $xmlWriter->endElement();
        $xmlWriter->endDocument();
        return $xmlWriter->outputMemory();
    }

    //消费
    private function sendConsumeRequest(string $topic, int $numOfMessages, string $consumer, int $waitSeconds = -1) {
        $body = null;
        $method = 'GET';
        $queryString = $this->generateConsumeQueryString($numOfMessages, $consumer, $waitSeconds);

        $resourcePath = 'topics/' . $topic . '/messages';
        $headers = $this->addRequiredHeaders($resourcePath, $body, $method, $queryString);
        $request = new Request($method, $resourcePath, $headers);
        try {
            $response = $this->client->send($request, [
                'query' => $queryString,
                'body' => $body,
                'timeout' => $this->requestTimeout,
                'connect_timeout' => $this->connectTimeout
            ]);
            $statusCode = $response->getStatusCode();
            $content = $response->getBody();

            if ($statusCode == 200) {
                return $this->parseConsumeXml($content);
            } else {
                //Log::error('MQ Consume Error ' . $statusCode, ["error" => '']);
                throw new MQException($statusCode, 'MQ Consume Error');
            }
        } catch (GuzzleException $exception) {
            if ($exception instanceof ClientException) {
                /** @var ClientException $exception */
                $code = $exception->getResponse()->getStatusCode();
                if ($code == 404) {
                    return [];
                }
            }
            //Log::error('MQ Consume Network Error', ["error" => $exception->getMessage()]);
            throw new MQException(500, 'MQ Consume Network Error');
        }

    }

    private function parseConsumeXml($content) {
        $res = [];
        $xmlReader = new XMLReader();

        $isXml = $xmlReader->XML($content);
        if ($isXml === FALSE) {
            throw new MQException(500, $content);
        }

        try {
            while ($xmlReader->read()) {
                if ($xmlReader->nodeType == XMLReader::ELEMENT
                    && $xmlReader->name == 'Message') {
                    $res [] = Message::fromXML($xmlReader);
                }
            }
            return $res;
        } catch (Exception $e) {
            throw new MQException(500, $e->getMessage());
        }
    }

    private function parseProduceXml($content) {
        $xmlReader = new XMLReader();
        $isXml = $xmlReader->XML($content);
        if ($isXml === FALSE) {
            throw new MQException(500, $content);
        }
        try {
            return Message::fromXML($xmlReader)->messageId;
        } catch (Exception $exception) {
            throw new MQException($exception->getCode(), $exception->getMessage(), $exception);
        }
    }
    
    private function generateConsumeQueryString(int $numOfMessages, string $consumer, int $waitSeconds = -1) {
        $params = array("numOfMessages" => $numOfMessages);
        $params["consumer"] = $consumer;
        if ($this->instanceId != NULL && $this->instanceId != "") {
            $params["ns"] = $this->instanceId;
        }
        if ($waitSeconds != NULL) {
            $params["waitseconds"] = $waitSeconds;
        }
        return http_build_query($params);
    }

    //Ack
    private function sendAckRequest(string $topic, array $receiptHandles, string $consumer) {
        $body = $this->generateAckBody($receiptHandles);
        $method = 'DELETE';
        $queryString = $this->generateAckQueryString($consumer);

        $resourcePath = 'topics/' . $topic . '/messages';
        $headers = $this->addRequiredHeaders($resourcePath, $body, $method, $queryString);
        $request = new Request($method, $resourcePath, $headers);
        try {
            $response = $this->client->send($request, [
                'query' => $queryString,
                'body' => $body,
                'timeout' => $this->requestTimeout,
                'connect_timeout' => $this->connectTimeout
            ]);
            $statusCode = $response->getStatusCode();

            if ($statusCode == 204) {
                return true;
            } else {
                //Log::error('MQ Ack Error ' . $statusCode, ["error" => '']);
                throw new MQException($statusCode, 'MQ Ack Error');
            }
        } catch (GuzzleException $exception) {
            //Log::error('MQ Ack Network Error', ["error" => $exception->getMessage()]);
            throw new MQException(500, 'MQ Ack Network Error');
        }
    }

    private function generateAckQueryString(string $consumer) {
        $params = array("consumer" => $consumer);
        if ($this->instanceId != null && $this->instanceId != "") {
            $params["ns"] = $this->instanceId;
        }
        return http_build_query($params);
    }

    private function generateAckBody(array $receiptHandles) {
        $xmlWriter = new XMLWriter;
        $xmlWriter->openMemory();
        $xmlWriter->startDocument("1.0", "UTF-8");
        $xmlWriter->startElementNS(NULL, Constants::RECEIPT_HANDLES, Constants::XML_NAMESPACE);
        foreach ($receiptHandles as $receiptHandle) {
            $xmlWriter->writeElement(Constants::RECEIPT_HANDLE, $receiptHandle);
        }
        $xmlWriter->endElement();
        $xmlWriter->endDocument();
        return $xmlWriter->outputMemory();
    }

    //通用
    private function addRequiredHeaders(string $resourcePath, $body, string $method, string $queryString, string $agent = '') {
        $headers = ['User-Agent' => $agent];
        if ($body != NULL) $headers['Content-Length'] = strlen($body);
        $headers['Date'] = gmdate("D, d M Y H:i:s \\G\\M\\T");
        $headers['Content-Type'] = 'text/xml';
        $headers['x-mq-version'] = '2015-06-06';

        $sign = $this->sign($resourcePath, $headers, $method, $queryString);
        $headers['Authorization'] = 'MQ ' . $this->accessId . ':' . $sign;

        return $headers;
    }

    private function sign(string $resourcePath, array $headers, string $method, string $queryString) {
        $contentMd5 = "";
        if (isset($headers['Content-MD5'])) {
            $contentMd5 = $headers['Content-MD5'];
        }
        $contentType = "";
        if (isset($headers['Content-Type'])) {
            $contentType = $headers['Content-Type'];
        }
        $date = $headers['Date'];
        $canonicalizedResource = $resourcePath;
        if ($queryString != NULL) {
            $canonicalizedResource .= "?" . $queryString;
        }
        if (0 !== strpos($canonicalizedResource, "/")) {
            $canonicalizedResource = "/" . $canonicalizedResource;
        }
        $tmpHeaders = array();
        foreach ($headers as $key => $value) {
            if (0 === strpos($key, 'x-mq-')) {
                $tmpHeaders[$key] = $value;
            }
        }
        ksort($tmpHeaders);

        $canonicalizedHeaders = implode(
            "\n",
            array_map(function ($v, $k) {
                return $k . ":" . $v;
            }, $tmpHeaders, array_keys($tmpHeaders))
        );

        $stringToSign = strtoupper($method) . "\n" . $contentMd5 . "\n" . $contentType . "\n" . $date . "\n" . $canonicalizedHeaders . "\n" . $canonicalizedResource;

        return base64_encode(hash_hmac("sha1", $stringToSign, $this->accessKey, true));
    }

    private function checkPropValid(array $properties) {
        foreach ($properties as $key => $value) {
            if ($key === NULL || $key == "" || $value === NULL || $value == "") {
                throw new MQException(400, "Message Properties is null or empty");
            }

            if ($this->isContainSpecialChar($key) || $this->isContainSpecialChar($value)) {
                throw new MQException(400, "Message's property can't contains: & \" ' < > : |");
            }
        }
    }

    private function isContainSpecialChar(string $str) {
        return strpos($str, "&") !== FALSE
            || strpos($str, "\"") !== FALSE || strpos($str, "'") !== FALSE
            || strpos($str, "<") !== FALSE || strpos($str, ">") !== FALSE
            || strpos($str, ":") !== FALSE || strpos($str, "|") !== FALSE;
    }
}
